#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include <stdbool.h>
#include "avl.h"

void avl_print(arbre tree,int N) {
   if (N <= 0) {
      N = 1;
   }
   if (NULL != tree) {
      avl_print(tree->child[1],N+2);
      for (int i=0;i<N;i++) {
         printf("    ");
      }
      printf("%d\n",tree->key);            
      avl_print(tree->child[0],N+2);
   } 
}

node* avl_desequilibre(arbre tree,clef key) {
   node* crt = tree;
   node* parent = NULL;
   int h_prec[2] = {0,0};

   while (NULL != crt && crt->key != key) {
      if (abs(crt->h[1]-crt->h[0]) == 2) {
         parent = crt;
         h_prec[0] = parent->h[0];
         h_prec[1] = parent->h[1];
         parent->h[key > parent->key] -= 1;
      }
      crt = crt->child[key > crt->key];
   }
   if (NULL != parent) {
      parent->h[0] = h_prec[0];
      parent->h[1] = h_prec[1];
   }
   return parent;
}

int avl_height(arbre tree) {
   if (NULL == tree) {
      return 0;
   } else {   
      return 1+fmax(avl_height(tree->child[0]),avl_height(tree->child[1]))+0.01;
   }
}

void avl_heights(arbre tree) {
   for (int i=0;i<2;i++) {
      if (NULL != tree->child[i]) {
         avl_heights(tree->child[i]);
         tree->h[i] = 1+fmax(tree->child[i]->h[0],tree->child[i]->h[1])+0.01;
      } else {
         tree->h[i] = 0;
      }
   }
}

void avl_height_node(node* nd) {
   for (int i=0;i<2;i++) {
      if (NULL != nd->child[i]) {
         nd->h[i] = 1+fmax(nd->child[i]->h[0],nd->child[i]->h[1])+0.01;
      } else {
         nd->h[i] = 0;
      }
   }
}

static node* avl_rot(node* P,bool right) {
   node* Q = NULL;
   if (NULL != P) {
      Q                = P->child[right];
      P->child[right]  = Q->child[!right];
      Q->child[!right] = P;
      avl_height_node(P);
      avl_height_node(Q);
   }
   return Q;
}

node* avl_rot_g(node* nd) {
   return avl_rot(nd,false);
}

node* avl_rot_d(node* nd) {
   return avl_rot(nd,true);
}

static node* avl_rot2(node* nd,bool dg) {
   nd->child[dg] = avl_rot(nd->child[dg],dg);
   return avl_rot(nd,!dg);
}

node* avl_rot_gd(node* nd) {
   return avl_rot2(nd,false);
}

node* avl_rot_dg(node* nd) {
   return avl_rot2(nd,true);
}

static void avl_rotation(node* nd,bool right) {
   node* new_nd = calloc(1,sizeof(node));
   new_nd->key = nd->key;
   new_nd->child[right] = nd->child[right];
   new_nd->child[!right] = nd->child[!right]->child[right];
   node* tmp = nd->child[!right]; 

   nd->key = nd->child[!right]->key;
   nd->child[right] = new_nd;
   nd->child[!right] = nd->child[!right]->child[!right];
   avl_height_node(new_nd);
   avl_height_node(nd);
   free(tmp);
}

void avl_rotation_g(node* nd) {
   avl_rotation(nd,false);
}

void avl_rotation_d(node* nd) {
   avl_rotation(nd,true);
}

static void avl_rotation2(node* nd,bool dg) {
   avl_rotation(nd->child[dg],dg);
   avl_rotation(nd,!dg);
}

void avl_rotation_gd(node* nd) {
   avl_rotation2(nd,false);
}

void avl_rotation_dg(node* nd) {
   avl_rotation2(nd,true);
}

void avl_reequilibre(node* nd) {
   switch(nd->h[1]-nd->h[0]) {
      case +2:
         switch(nd->child[1]->h[1] - nd->child[1]->h[0]) {
            case +1: avl_rotation_g(nd); break;
            case -1: avl_rotation_dg(nd);
         }
         break;
      case -2:
         switch(nd->child[0]->h[1] - nd->child[0]->h[0]) {
            case -1: avl_rotation_d(nd); break;
            case +1: avl_rotation_gd(nd);
         }
   }
}

void avl_insert(clef key,arbre* tree,arbre* deseq) {
   if (NULL == *tree) {
      *tree = calloc(1,sizeof(node));
      (*tree)->key = key;
      *deseq = NULL;
   } else {
      avl_insert(key,&((*tree)->child[key > (*tree)->key]),deseq);
      if (NULL == *deseq) {
         (*tree)->h[key > (*tree)->key] 
             = 1+fmax((*tree)->child[key > (*tree)->key]->h[0],
                      (*tree)->child[key > (*tree)->key]->h[1]);
      } 
      if (abs((*tree)->h[1]-(*tree)->h[0]) == 2 && NULL == *deseq) {
         *deseq = *tree;
      }
   }
}

